// Copyright stuff goes here when done (ask Terence)

// include kernel i2c stuff
#define  __NO_VERSION__

#include "nv-misc.h"
#include "os-interface.h"
#include "nv-linux.h"

#if (defined(CONFIG_I2C) || defined(CONFIG_I2C_MODULE))

#if defined(KERNEL_2_4)
extern int i2c_add_adapter (struct i2c_adapter *) __attribute__ ((weak));
extern int i2c_del_adapter (struct i2c_adapter *) __attribute__ ((weak));
#define NV_WEAK_SYMBOL_PRESENT(x) (x != NULL)
#endif

// static local functions used for the algorithm
static int nv_i2c_algo_xfer(struct i2c_adapter *adap, struct i2c_msg msgs[], int num);
static int nv_i2c_algo_control(struct i2c_adapter *adap, unsigned int dummy1, unsigned long dummy2);
static u32 nv_i2c_algo_functionality(struct i2c_adapter *adap);

static struct i2c_algorithm nv_i2c_algo = {
    .master_xfer      = nv_i2c_algo_xfer,
    .smbus_xfer       = NULL,
    .algo_control     = nv_i2c_algo_control,
    .functionality    = nv_i2c_algo_functionality,
};

static int nv_i2c_algo_xfer(struct i2c_adapter *adap, struct i2c_msg msgs[], int num)
{
    // get a gpu to do i2c with
    void * nv_adap = (struct nv_i2c_algo_data *)adap->algo_data;
    void * kern_adap = (void *)adap;
    int i = 0;
    RM_STATUS rmStatus = RM_OK;

    for(i = 0; ((i < num) && (rmStatus == RM_OK)); i++)
    {
        // verify that the flags make sense
        if(
            (msgs[i].flags & I2C_M_TEN         ) ||
            (msgs[i].flags & I2C_M_NOSTART     ) ||
            (msgs[i].flags & I2C_M_REV_DIR_ADDR))
        {
            // we don't support ten bit addresses, no start sending receiving
            // or reverse direction addressing currently.
            return (-EPROTONOSUPPORT);
        }
            
        if(msgs[i].flags & I2C_M_RD)
        {
            rmStatus = rm_i2c_read_buffer_only((nv_state_t *)nv_adap,
                                            kern_adap,
                                    (U008) (msgs[i].addr & 0xff),
                                    (U032)  msgs[i].len & 0xffffUL,
                                    (U008 *)msgs[i].buf);
        }
        else
        {
            rmStatus = rm_i2c_write_buffer_only((nv_state_t *)nv_adap,
                                             kern_adap,
                                     (U008) (msgs[i].addr & 0xff),
                                     (U032)  msgs[i].len & 0xffffUL, 
                                     (U008 *)msgs[i].buf);
        }
    }        

    if(rmStatus != RM_OK)
        return (-EIO);
    else
        return num;
}

static int nv_i2c_algo_control(struct i2c_adapter *adap, unsigned int dummy1, unsigned long dummy2)
{
    return 0;
}

static u32 nv_i2c_algo_functionality(struct i2c_adapter *adap)
{
    return I2C_FUNC_I2C;
}

static int nv_i2c_adapter_register_client(struct i2c_client *);
static int nv_i2c_adapter_unregister_client(struct i2c_client *);

#if defined(KERNEL_2_4)

#if defined(NV_I2C_ADAPTER_INC_USE_PRESENT)
// defines for adapter implemenation
static void nv_i2c_adapter_inc_use(struct i2c_adapter * pAdapter);
static void nv_i2c_adapter_dec_use(struct i2c_adapter * pAdapter);
#endif

struct i2c_adapter nv_i2c_adapter_prototype = {
    .name              = "NVIDIA I2C Device",
    .id                = I2C_HW_B_RIVA,
    .algo              = &nv_i2c_algo,
    .algo_data         = NULL,
#if defined(NV_I2C_ADAPTER_INC_USE_PRESENT)
    .inc_use           = nv_i2c_adapter_inc_use,
    .dec_use           = nv_i2c_adapter_dec_use,
#endif
    .client_register   = nv_i2c_adapter_register_client,
    .client_unregister = nv_i2c_adapter_unregister_client,
    .data              = NULL,
};

#if defined(NV_I2C_ADAPTER_INC_USE_PRESENT)
static void nv_i2c_adapter_inc_use(struct i2c_adapter * pAdapter)
{
    MOD_INC_USE_COUNT;

    return;
}

static void nv_i2c_adapter_dec_use(struct i2c_adapter * pAdapter)
{
    MOD_DEC_USE_COUNT;

    return;
}
#endif

static int nv_i2c_adapter_register_client(struct i2c_client * pClient)
{
    MOD_INC_USE_COUNT;

    return 0;
}

static int nv_i2c_adapter_unregister_client(struct i2c_client * pClient)
{
    MOD_DEC_USE_COUNT;

    return 0;
}


#else // defined(KERNEL_2_4)

struct i2c_adapter nv_i2c_adapter_prototype = {
    .owner     = THIS_MODULE,
    .algo      = &nv_i2c_algo,
    .algo_data = NULL,
    .name      = "NVIDIA I2C Device",
    .client_register   = nv_i2c_adapter_register_client,
    .client_unregister = nv_i2c_adapter_unregister_client,
};

static int nv_i2c_adapter_register_client(struct i2c_client * pClient)
{
    try_module_get(THIS_MODULE);
    return 0;
}

static int nv_i2c_adapter_unregister_client(struct i2c_client * pClient)
{
    module_put(THIS_MODULE);
    return 0;
}

#endif // defined(KERNEL_2_4)

void * NV_API_CALL os_i2c_add_adapter(void * Opaque_pNvAlgoData)
{
    RM_STATUS rmStatus = RM_OK;
    struct i2c_adapter *pI2cAdapter = NULL;
    int osstatus = 0;

#if defined(KERNEL_2_4)
    if (!NV_WEAK_SYMBOL_PRESENT(i2c_add_adapter))
        return NULL;
    if (!NV_WEAK_SYMBOL_PRESENT(i2c_del_adapter))
        return NULL;
#endif

    // get a i2c adapter
    rmStatus = os_alloc_mem((void **)&pI2cAdapter,sizeof(struct i2c_adapter));

    if(rmStatus != RM_OK)return NULL;

    // fill in with default structure
    os_mem_copy((void *)pI2cAdapter, (void *)&nv_i2c_adapter_prototype, sizeof(struct i2c_adapter));

    // add our data to the structure
    pI2cAdapter->algo_data = Opaque_pNvAlgoData;

    // attempt to register with the kernel
    osstatus = i2c_add_adapter(pI2cAdapter);

    if(osstatus)
    {
        // free the memory and NULL the ptr
        os_free_mem(pI2cAdapter);

        pI2cAdapter = NULL;
    }

    return ((void *)pI2cAdapter);
}

BOOL NV_API_CALL os_i2c_del_adapter(void * Opaque_pI2cAdapter)
{
    // this function should not be called if the i2c structure is not actually registered already.

    struct i2c_adapter * pI2cAdapter = (struct i2c_adapter *)Opaque_pI2cAdapter;
    int osstatus = 0;
    BOOL wasReleased = FALSE;

#if defined(KERNEL_2_4)
    if (!NV_WEAK_SYMBOL_PRESENT(i2c_add_adapter))
        return FALSE;
    if (!NV_WEAK_SYMBOL_PRESENT(i2c_del_adapter))
        return FALSE;
#endif

    // can't release a NULL ptr, return FALSE
    if(!pI2cAdapter) return FALSE;

    // attempt release with the OS
    osstatus = i2c_del_adapter(pI2cAdapter);

    if(!osstatus)
    {
        os_free_mem(pI2cAdapter);
        wasReleased = TRUE;
    }

    return wasReleased;
}

#else // (defined(CONFIG_I2C) || defined(CONFIG_I2C_MODULE))

BOOL NV_API_CALL os_i2c_del_adapter(void * Opaque_pI2cAdapter)
{
    return FALSE;
}

void * NV_API_CALL os_i2c_add_adapter(void * Opaque_pNvAlgoData)
{
    return NULL;
}

#endif // (defined(CONFIG_I2C) || defined(CONFIG_I2C_MODULE))
